#!/bin/sh
# Collation Discover Server setup script.
#
# Sets up the discover anchor server environment:
# - verifies the Java version
# - creates and/or verifies directory structure:
#	- coll, lib, tmpjars, log, etc
#
# See startup.sh to start the server.

# not really constants, just OS-specific stuff
setosconstants()
{
	case "$OS" in
	SunOS)
		SOLARISNONL="\c"
		MYSHELL=/bin/ksh
		JETTYTMP=/var/tmp
		HASCHKCONFIG=no
		UNZIP=/bin/unzip
		ARCH=`uname -p`
		;;
	Linux)
		LINUXNONL="-n"
		MYSHELL=/bin/bash
		JETTYTMP=/tmp
		HASCHKCONFIG=yes
		UNZIP=/usr/bin/unzip
		ARCH=`uname -m`
		;;
	AIX)
		SOLARISNONL="\c"                  
		MYSHELL=/bin/ksh
		JETTYTMP=/tmp
		HASCHKCONFIG=no
		UNZIP=/usr/local/bin/unzip
		ARCH=`uname -p`
		;;			
	esac
}

if [ $# -ne 1 ]
then
        echo "Usage:- $0 java_version"
        echo "java_version should be of the form: 1.x.y, e.g. 1.4.2"
        exit 1
fi

VERS=$1

ROOT=`pwd`/coll4.0

if [ ! -d coll4.0 ]
then
  mkdir coll4.0
  if [ $? -ne 0 ]
  then
    echo "Error: unable to create coll4.0 directory"
    exit 1
  fi
  chmod 700 coll4.0
fi

cd coll4.0
if [ $? -ne 0 ]
then
  echo "Error: unable to connect to coll4.0 directory"
  exit 1
fi

for i in tmpjar log bin etc etc/templates etc/jboss lib lib/ms lib/ms/gateway lib/ms/gateway/share external external/jdk external/nmap external/nmap/SunOS external/nmap/Linux
do
  if [ ! -d $i ]
  then
    mkdir $i
    if [ $? -ne 0 ]
    then
      echo "Error: unable to create $i directory"
      exit 1
    fi
  fi
done


OS=`uname -s`
setosconstants
JDK_VERSION=1.5.0
JAVA_HOME=$ROOT/external/jdk-${JDK_VERSION}-${OS}-${ARCH}

#different java directories structure on AIX
if [ "$OS" = "AIX" ] 
then
    JAVA_HOME=$JAVA_HOME/jre
fi

if [ -d $JAVA_HOME ]
then
  export JAVA_HOME
else
  echo "java not found, need to deploy it"
  # return code to signal bad java version
  exit 42
fi

PATH=$JAVA_HOME/bin:$HOME/coll4.0/bin:$PATH
export PATH

java -version > /dev/null 2>&1
if [ $? -ne 0 ]
then
  echo "Error: java not found"
  # return code to signal bad java version
  exit 42
fi

FOUND=`java -version 2>&1 | grep 'java version' | sed "s/^[^\"]*//g" | cut -c2-6`

if [ x$VERS != x$FOUND ]
then
  echo "Error: java version incompatible.  Found: $FOUND.  Required: $VERS."
  exit 42
fi

echo "setup passed"
