#!/bin/sh
#
# Copyright 2001-2003 Collation Inc. All Rights Reserved
#
# This software is the proprietary information of Collation Inc.
# Use is subject to license terms.
#
# Collation Discover Anchor Server startup script.
#
# starts the discover anchor server:
# - sets class path
# - verifies that no CollationServer is already running
# - tests that the server is runnable
# - starts it, redirecting output to a unique log file

# must be run in the coll4.0 sub-directory

# not really constants, just OS-specific stuff
setosconstants()
{
	case "$OS" in
	SunOS)
		SOLARISNONL="\c"
		MYSHELL=/bin/ksh
		JETTYTMP=/var/tmp
		HASCHKCONFIG=no
		UNZIP=/bin/unzip
		ARCH=`uname -p`
		;;
	Linux)
		LINUXNONL="-n"
		MYSHELL=/bin/bash
		JETTYTMP=/tmp
		HASCHKCONFIG=yes
		UNZIP=/usr/bin/unzip
		ARCH=`uname -m`
		;;
	AIX)
		SOLARISNONL="\c"                  
		MYSHELL=/bin/ksh
		JETTYTMP=/tmp
		HASCHKCONFIG=no
		UNZIP=/usr/local/bin/unzip
		ARCH=`uname -p`
		;;		
	esac
}

if [ $# -ne 2 ]
then
        echo "Usage:- $0 port_num ip_address"
        exit 1
fi

PORT=$1
IP=$2

echo "classpath=$CLASSPATH"
FOO=.
for i in lib/anchor.jar lib/platform-model.jar lib/discover-sensors.jar lib/platform-snmp.jar lib/discover-sensors-dep.jar  
do
FOO=$FOO:$i
done
for i in lib/anchor_*.jar
do
	if [ ! -f $i ]
	then
		break
	fi
FOO=$FOO:$i
done
CLASSPATH=$FOO:$CLASSPATH
export CLASSPATH
echo "classpath=$CLASSPATH"
pwd

ROOT=`pwd`
OS=`uname -s`
setosconstants
JDK_VERSION=1.5.0
JAVA_HOME=$ROOT/external/jdk-${JDK_VERSION}-${OS}-${ARCH}

#different java directories structure on AIX
if [ "$OS" = "AIX" ] 
then
    JAVA_HOME=$JAVA_HOME/jre
fi

if [ -d $JAVA_HOME ]
then
  export JAVA_HOME
else
  echo "Error: unable to locate java in $JAVA_HOME"
  exit 1
fi

PATH=$JAVA_HOME/bin:$HOME/coll4.0/bin:$PATH
export PATH

java com.collation.discover.anchor.CollationServer -version > tmp$$.log 2>&1
if [ $? -ne 0 ]
then
  cat tmp$$.log
  rm tmp$$.log
  echo "Error: unable to start Collation Server"
  exit 1
fi

# See if we need to run 64 bit mode
if grep '^com\.collation\.64bit=true' $ROOT/etc/collation.properties >/dev/null
then
    BITS="-d64"
else
    BITS=
fi

fgrep -i copyright tmp$$.log > /dev/null 
if [ $? -ne 0 ]
then
  cat tmp$$.log
  rm tmp$$.log
  echo "Error: unable to start Collation Server (b)"
  exit 1
fi
rm tmp$$.log

rm -f tmpjar/*.jar

MYHOST=`hostname`
LOGFILE="-Dcom.collation.LogFile=log/anchor.$MYHOST.log"
OUTLOG=`date +"CollationServer.%d%m%y.%H%M%S.$MYHOST.out"`

#The "-server" option not supported by IBM java
java $BITS -DCOLLATION -Dcom.collation.home=`pwd` $LOGFILE com.collation.discover.anchor.CollationServer -d 1 -p $PORT -i $IP > log/$OUTLOG 2>&1 &

if [ $? -ne 0 ]
then
  cat log/$OUTLOG
  echo "Error: unable to start Collation Server"
  exit 1
fi
echo "server started, `date`"
