/*
 * $Id: OracleAgent.java,v 1.131.4.1 2006/06/30 00:34:00 cmvc Exp $
 *
 * Copyright 2001-2005 Collation Inc. All Rights Reserved
 *
 * This software is the proprietary information of Collation Inc.
 * Use is subject to license terms.
 *
 */
package com.collation.discover.agent.app.db.oracle;

import com.collation.platform.configuration.Configurable;
import com.collation.platform.configuration.ConfigurationException;
import com.collation.platform.ip.IpUtils;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.Driver;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.Properties;

import com.collation.discover.agent.AbstractAgent;
import com.collation.discover.agent.AgentException;
import com.collation.discover.agent.app.appdescriptor.AppDescriptorAgentDelegate;
import com.collation.discover.agent.messages.DiscoverAgentLocalizedMessages;
import com.collation.discover.result.Result;
import com.collation.discover.result.app.db.oracle.OracleResult;
import com.collation.discover.seed.Seed;
import com.collation.discover.seed.app.db.oracle.OracleSeed;
import com.collation.platform.logger.LogFactory;
import com.collation.platform.logger.Logger;
import com.collation.platform.model.AttributeNotSetException;
import com.collation.platform.model.Configuration;
import com.collation.platform.model.discovery.agent.OracleAgentConfiguration;
import com.collation.platform.model.topology.app.ConfigFile;
import com.collation.platform.model.topology.app.AppConfig;
import com.collation.platform.model.topology.app.ProcessPool;
import com.collation.platform.model.topology.app.SoftwareModule;
import com.collation.platform.model.topology.app.db.oracle.OracleBackgroundProcess;
import com.collation.platform.model.topology.app.db.oracle.OracleControlFile;
import com.collation.platform.model.topology.app.db.oracle.OracleDBLink;
import com.collation.platform.model.topology.app.db.oracle.OracleDataFile;
import com.collation.platform.model.topology.app.db.oracle.OracleDatabase;
import com.collation.platform.model.topology.app.db.oracle.OracleInitValue;
import com.collation.platform.model.topology.app.db.oracle.OracleInstance;
import com.collation.platform.model.topology.app.db.oracle.OracleListener;
import com.collation.platform.model.topology.app.db.oracle.OracleRedoLogFile;
import com.collation.platform.model.topology.app.db.oracle.OracleSGAValue;
import com.collation.platform.model.topology.app.db.oracle.OracleSchema;
import com.collation.platform.model.topology.app.db.oracle.OracleModule;
import com.collation.platform.model.topology.app.db.oracle.OracleSchemaDetail;
import com.collation.platform.model.topology.app.db.oracle.OracleServer;
import com.collation.platform.model.topology.app.db.oracle.OracleServerProcess;
import com.collation.platform.model.topology.app.db.oracle.OracleTableSpace;
import com.collation.platform.model.topology.core.LogicalContent;
import com.collation.platform.model.topology.net.BindAddress;
import com.collation.platform.model.topology.net.IpAddress;
import com.collation.platform.model.topology.sys.OperatingSystem;
import com.collation.platform.model.topology.sys.RuntimeProcess;
import com.collation.platform.security.auth.AuthManager;
import com.collation.platform.security.auth.DbAuth;
import com.collation.platform.session.SessionClient;
import com.collation.platform.session.SessionClientFactory;
import com.collation.platform.util.InterMessageUtils;
import com.collation.platform.util.ModelFactory;
import com.collation.platform.util.Props;
import com.collation.platform.ip.ScopedProps;
import com.collation.platform.util.URIUtils;
import com.collation.platform.util.SecurityUtils;
import com.collation.platform.os.Os;
import com.collation.platform.os.OsException;
import com.collation.platform.os.OsCmd;
import com.collation.platform.os.OsFactory;

/**
 * 
 * OracleAgent requires: - oracle to be up, with a listener running - a ssh
 * connection, to grep for: - sid info from oratab - port info from lsof - able
 * to connect via jdbc - needs select_catalog_role privileges
 * 
 * This agent is not yet finished. Left to do: FIX THIS - this code assumes a
 * 1-1 mapping betweeb db & instance, and sid & instance - create a
 * multi-instance/db system, with sid aliases and see how it goes - should be
 * further customized for Oracle8 - left to discover: segments
 * 
 */
public class OracleAgent extends AbstractAgent implements Configurable {
    private Logger log_ = null;
    private static boolean driverLoaded = false;
    private ArrayList authList_ = null;
    private Driver driver_ = null;

    private OracleResult result_;
    private SessionClient session_ = null;

    private transient HashMap contentMap = new HashMap();
    private HashMap badEndpoints_ = new HashMap();

    private static final int DEFAULT_PORT = 1521;

    String hostName_ = null;
    String ipName_ = null;
    OracleSeed oraSeed_ = null;
    Os os_ = null;
    boolean isWindows_ = false;
    boolean isOpenVms_ = false;
    
    boolean useConnectorThreads_ = true;
    long connThreadTimeout_ = 10000;

    HashMap suppressPorts_ = new HashMap();
    HashMap suppressSIDs_ = new HashMap();

    IpAddress primaryIp_;
    int[] seedPorts_;
    static volatile HashMap psCache_ = new HashMap(100);

    // sid/port <-> conn
    private transient HashMap connPool = new HashMap();

    private transient ArrayList instanceList_ = null;

    private transient HashMap serverMap_ = new HashMap();
    private transient HashMap sidMap_ = new HashMap();

    // listener name <-> port list array
    private transient HashMap portList = null;

    // plus session info

    // version of Oracle, either 8 or 9
    private int oracleVersion_ = 0;
    // Added for dblink
    private String oraHome = ""; //$NON-NLS-1$

    // how many tables to select
    private static final int TABLE_MAX = Props.getTableLimit();

    private OracleAgentConfiguration configuration_;

    /**
     * Return the interface list
     */
    public Result getResult() {
        return result_;
    }

    /**
     * Set the discover result.
     */
    public void setResult(Result result) {
        result_ = (OracleResult) result;
    }

    /**
     * Return a null set of required jars.
     */
    public String[] getRequiredJars() {
        return new String[] { "classes12.jar" }; //$NON-NLS-1$
    }

    /**
     * Try to create a session immediately from the seed. The session must have
     * privileges to see the oratab and tnsnames.ora files.
     */
    public void setSeed(Seed seed) {
        try {
            log_ = LogFactory.getLogger("OracleAgent"); //$NON-NLS-1$
            oraSeed_ = (OracleSeed) seed;

        } catch (Exception ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.0"), ex); //$NON-NLS-1$
        }
    }

    /**
     * How Oracle info is discovered:
     * 
     * Instances are found via ps info (ora_pmon_<sid>). This contains the SID
     * and oracle home. This is used to find the port from tnsnames.ora. Then
     * JDBC must be able to connect, as all further discovery is done via SQL.
     * 
     * In general, each each method follows the result model, creating a single
     * array/object and filling in member model objects with calls to other
     * oracle discovery methods.
     * 
     * NOTE:
     * OpenVMS tasknames are in format ORA_<sid>_PMON.
     *
     * 
     */
    public void discover() throws AgentException, InterruptedException {
        int dbFound = 0;

        //Load the JDBC driver if it is not loaded already
        try {
            synchronized (OracleAgent.class){
                if (driverLoaded == false){
                    DriverManager.registerDriver((java.sql.Driver) Class.forName("oracle.jdbc.driver.OracleDriver").newInstance()); //$NON-NLS-1$
                    driverLoaded = true;
                }
            }
        }
        catch (Exception ex){
            log_.debug("Could not load jdbc driver",ex); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.2", //$NON-NLS-1$
                    new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex);
        }

        try {
            instanceList_ = new ArrayList();

            session_ = SessionClientFactory.createSessionClient(oraSeed_.getIpAddress());

            if (session_ == null) {
                throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.1"); //$NON-NLS-1$
            }
            os_ = OsFactory.newOs(session_);
            isWindows_ = (os_ instanceof com.collation.platform.os.WindowsOs);
            isOpenVms_ = (os_ instanceof com.collation.platform.os.OpenVmsOs);

            primaryIp_ = oraSeed_.getIpAddress();
            seedPorts_ = oraSeed_.getPortList();

            useConnectorThreads_ = ScopedProps.getBooleanProperty("com.collation.discover.agent.OracleAgent.useConnectorThreads",primaryIp_.getDotNotation(),true);
            connThreadTimeout_ = ScopedProps.getIntProperty("com.collation.discover.agent.OracleAgent.connectionThreadTimeout",primaryIp_.getDotNotation(),10000); //$NON-NLS-1$
            String suppressPorts = ScopedProps.getStringProperty("com.collation.discover.agent.OracleAgent.suppressPorts",primaryIp_.getDotNotation());
            String suppressSIDs = ScopedProps.getStringProperty("com.collation.discover.agent.OracleAgent.suppressSIDs",primaryIp_.getDotNotation());
            if (suppressPorts != null){
                String [] supp_ports = suppressPorts.split(","); //$NON-NLS-1$
                for (int i=0; i< supp_ports.length; i++){
                    try {
                        suppressPorts_.put(Integer.valueOf(supp_ports[i]),Integer.valueOf(supp_ports[i]));
                        log_.debug("Added port to suppress list: " + supp_ports[i]); //$NON-NLS-1$
                    }
                    catch (NumberFormatException nfex){
                    }
                }
            }
            if (suppressSIDs != null){
                String[] supp_sids = suppressSIDs.split(","); //$NON-NLS-1$
                for (int i=0; i<supp_sids.length; i++){
                    suppressSIDs_.put(supp_sids[i].trim().toLowerCase(),supp_sids[i]);
                    log_.debug("Added sid to suppress list: " + supp_sids[i]); //$NON-NLS-1$
                }
            }
            // convert ip address to a host name, which Oracle uses
            hostName_ = InetAddress.getByName(primaryIp_.getDotNotation()).getHostName();
            ipName_ = primaryIp_.getDotNotation();

            result_ = new OracleResult();

            ArrayList ia = new ArrayList();

            OracleServer[] servers = findServers(ia);

            for (int i = 0; i < servers.length; i++) {
                OracleServer server = servers[i];

                try {
                    findDatabases(server);
                    dbFound++;
                } catch (AgentException ex) {
                    log_.debug("Error finding databases:" + ex + ", continuing..."); //$NON-NLS-1$ //$NON-NLS-2$
                }

                server.setHost(os_.getComputerSystem());
            }

            // instances are set separately to break up result into
            // transactionaly manageable chunks

            if (instanceList_.size() > 0) {
                OracleInstance[] oarray = new OracleInstance[instanceList_.size()];
                result_.setOracleInstances((OracleInstance[]) instanceList_.toArray(oarray));
            }

            result_.setOracleServers(servers);
            discoverAppDescriptors();
        } catch (InterruptedException ex) {
            throw ex;
        } catch (AgentException ex) {
            throw ex;
        } catch (Exception ex) {
            log_.debug("General Exception", ex); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.2", //$NON-NLS-1$
                    new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex);
        } finally {
            freeConnections();
            session_ = null;
        }

        if (dbFound == 0) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.3")); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.3"); //$NON-NLS-1$
        }
        log_.debug("oracleagent, discover returning..."); //$NON-NLS-1$
    }

    private static final String oratabs[] = { "/etc/oratab", "/var/opt/oracle/oratab", "/var/opt/oratab" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

    /**
     * Finds all servers on the current machine, by looking at the oratab file.
     * It's assumed that is found at either /etc, or /var/opt/oracle. Here's
     * some example entries:
     * 
     * orcl9i:/opt/oracle9.2:Y OEMREP:/opt/oracle9.2:N # comments
     * OLD_SID:/bogus/directory:?
     * 
     * It's assumes that each oracle home represents a separate server. This
     * only sets the config file to the oracle home directory. The other members
     * will be filled in later.
     * 
     * If no servers are found, an exception is thrown.
     * 
     */
    public OracleServer[] findServers(ArrayList ia) throws AgentException, InterruptedException {
        String res = null;
        Vector v = new Vector(5);
        OracleServer server;

        try {
            // look at all listeners homes to get server list
            getServersFromPortList(v, ia);
        } catch (Exception ex) {
            // either an IO or an authentication error, which is fatal
            log_.debug("findServers: finding servers", ex); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.2", new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleServer[] oarray = new OracleServer[v.size()];
        return (OracleServer[]) v.toArray(oarray);
    }

    /**
     * It's possible to find the oracle homes from the tnslsnr ps output,
     * assuming the full path exists there.
     */
    private void getServersFromPortList(Vector v, ArrayList ia) throws AgentException, AttributeNotSetException {
        HashMap map = new HashMap(10);
        if (portList == null) {
            portList = getPortList();
        }

        boolean found = false;

        // iterate over all listeners found via ps,
        for (Iterator iter = portList.keySet().iterator(); iter.hasNext();) {
            String l = (String) iter.next();
            log_.debug("Checking " + l); //$NON-NLS-1$
            int i = 0;
            if (isWindows_){
                i = l.toLowerCase().lastIndexOf("bin\\"); //$NON-NLS-1$
            } 
            else if (isOpenVms_) {
    	        // Finding the oracle home dir on OpenVms is a bit different.  First,
    	        // we will use the PID to create a table name.  The table name will
    	        // then be searched for the logical name ARG0.  ARGO contains the 
    	        // path of the tnslsnr program in following format:
    	        //      system$drive:[oracle_home_dir.NETWORK]tnslsnr
    	        // From this, the oracle home directory can be obtained.
    	        Long pid = new Long(oraSeed_.getProcess().getPID());
    	        // Since pid is stored as a long and OpenVMS refers to pids in hex,
    	        // we need to convert it.
    	        String hexpid = Long.toHexString(pid.longValue());
    	        // The OpenVMS table name is format of TNS_xxxxxxxx where
    	        // xxxxxxxx is the pid number in hex
    	    	int numLeadingZeroes = 8 - hexpid.length();
    	    	if (numLeadingZeroes > 0) {
    	    		StringBuffer leadingZeroes = new StringBuffer();
    	    		for(int j=0; j<numLeadingZeroes;j++) {
    	    			leadingZeroes.append("0"); //$NON-NLS-1$
    	    		}
    	    		hexpid = leadingZeroes.toString() + hexpid;
    	    	}
    	    	String tableName = "TNS_" + hexpid.toUpperCase(); //$NON-NLS-1$
    	    	log_.debug("OracleAgent: OpenVMS table name is " + tableName); //$NON-NLS-1$
    	    	String cmd = "show logical /table=" + tableName + " ARG0"; //$NON-NLS-1$ //$NON-NLS-2$
    	    	try {
    				String res = session_.executeCommand(cmd);
    				if (res != null && res.length() > 0) {
    					Pattern pattern = Pattern.compile("= .(\\S+).NETWORK.*"); //$NON-NLS-1$
    					Matcher matcher = pattern.matcher(res);
    					String openVmsOracleHome = null;
    					if (matcher.find()) {
    						// Add bracket at end to complete the path
    						openVmsOracleHome = "/" + matcher.group(1) + "]"; //$NON-NLS-1$ //$NON-NLS-2$
    					}
    					log_.debug("OracleAgent: OpenVMS Oracle_Home is " + openVmsOracleHome); //$NON-NLS-1$
    					// To retrofit with existing code, set variable l to the oracle home
    					// and set to i to a positive value to indicate that we found something
    					l = openVmsOracleHome;
    					i = 1;
    				} 
    			} catch (Exception e) {
    				throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.27", //$NON-NLS-1$
                            new Object[]{cmd, InterMessageUtils.getIntactMessage(e)});
    			} 
    	    	
    	    } else {
    	        i = l.lastIndexOf("bin/"); //$NON-NLS-1$
    	    }
    	    if (i > 0) {
            	// OpenVMS finds the oracle home a different way (see above)
                if (!isOpenVms_) {
                    // found a match
                    l = l.substring(0, i).trim();
                }
                log_.debug("Found " + l); //$NON-NLS-1$
                if (map.get(l) != null) {
                    continue; // already found this home
                }
                map.put(l, l); // it's a new home, make a server from it

                OracleServer server = (OracleServer) ModelFactory.newInstance(OracleServer.class);
                ConfigFile cf = (ConfigFile) ModelFactory.newInstance(ConfigFile.class);
                cf.setURI(URIUtils.newURI(primaryIp_.getDotNotation(), l).toString());
                String str = l;
                if(isOpenVms_ && str.length() > 0) {
                	// omit '/' from path for OpenVMS since it uses a different
                	// directory format
                	str = str.substring(1);
                }

                if (str.length() > 1 && str.charAt(str.length() - 1) == '/') {
                    str = str.substring(0, str.length() - 1);
                }
                cf.setRealFile(str);
                cf.setFixedPath(str);
                server.setConfigFile(cf);

                try {
                    OracleInstance[] instances = findInstances(server);
                    // disabled to break result up
                    // server.setInstances(instances);

                    OracleListener[] listeners = findListener2(server);
                    server.setListeners(listeners);

                } catch (Exception ex) {
                    // either an IO or an authentication error, which is fatal
                    log_.debug("getServersFromPortList: finding servers", ex); //$NON-NLS-1$
                    continue; // not a real server
                }

                v.add(server);
                found = true;
            }
        }
    
        if (!found) {
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.4"); //$NON-NLS-1$
        }

    }

    // return true iff directory exists - no errors
    private boolean statDirectory(String dir) {
        try {
            // 1st, just check to see if the directory exists
            session_.executeCommand("test -d " + dir); //$NON-NLS-1$
        } catch (Exception ex) {
            return false;
        }
        return true;
    }

    /**
     * an alternate way of finding SIDS on windows
     */
    private String[] searchForSidsInRegistry(){
	ArrayList sids = new ArrayList();	
	Pattern p;
	Matcher m;
	String enabled =""; //$NON-NLS-1$
	String registrySearchRegexes = ""; //$NON-NLS-1$

        try{
	    enabled = ScopedProps.getStringProperty("com.collation.discover.agent.OracleAgent.searchWindowsRegistry",os_.getSession().getHost()); //$NON-NLS-1$
        }
        catch (OsException ex){
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.28"), ex); //$NON-NLS-1$
            enabled = null;
        }	
        if (enabled == null){
	    log_.debug("searchForSidsInRegistry: is disabled"); //$NON-NLS-1$
	    return (String [])sids.toArray(new String[0]);
        }	
        if (enabled.indexOf("false") >= 0){ //$NON-NLS-1$
	    log_.debug("searchForSidsInRegistry: is false"); //$NON-NLS-1$
	    return (String [])sids.toArray(new String[0]);
        }

        // Get a semicolon separated list of regexes to apply to the registry
        // key names to hopefully find some sids
        try{
            registrySearchRegexes = ScopedProps.getStringProperty("com.collation.discover.agent.OracleAgent.registrySearchRegexes",os_.getSession().getHost()); //$NON-NLS-1$
        }
        catch (OsException ex){
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.28"), ex); //$NON-NLS-1$
            registrySearchRegexes = null;
        }
        if (registrySearchRegexes == null){
            log_.debug("seachForSidsInRegistry: searching enabled but no regexes specified"); //$NON-NLS-1$
	    return (String [])sids.toArray(new String[0]);
        }         
        String[] regexes = registrySearchRegexes.split(";"); //$NON-NLS-1$
        try{
	    String[] registry_info = os_.queryRegistry("HKLM\\SOFTWARE\\ORACLE").split("\n"); //$NON-NLS-1$ //$NON-NLS-2$
            log_.debug("searchForSidsInRegistry: registry_info is " + registry_info.length + " lines long"); //$NON-NLS-1$ //$NON-NLS-2$
            for (int i = 0; i < regexes.length; i++){
                log_.debug("searchForSidsInRegistry: applying " + regexes[i]); //$NON-NLS-1$
                for (int j = 0; j < registry_info.length; j++){
                    p = Pattern.compile(regexes[i],Pattern.CASE_INSENSITIVE);
                    m = p.matcher(registry_info[j]);
                    if (m.find()){
			log_.debug("seachForSidsInRegistry: adding sid canidate " + m.group(1)); //$NON-NLS-1$
                        sids.add(m.group(1));
                    }
                }
            }
        }
        catch (Exception ex){
               log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.29"), ex); //$NON-NLS-1$
        }
        return (String [])sids.toArray(new String[0]); 
    } 

    /**
     * Finds all instances associated with the given server, by looking at the
     * oratab file. It's assumed that is found at either /etc, or
     * /var/opt/oracle. Here's some example entries:
     * 
     * orcl9i:/opt/oracle9.2:Y OEMREP:/opt/oracle9.2:N
     * 
     * Discovery of instance specific info, such as processes and SGA is
     * deferred until findDatabases, to optimize on sharing a connection.
     * 
     * If no instances are found, an exception is thrown.
     * 
     */
    public OracleInstance[] findInstances(OracleServer parent) throws AgentException, InterruptedException {
        Vector v = new Vector(5);
        String oracle_home = null;
        String oracle_sid;
        String oracle_dbname;
        boolean found = false;
        int i;

        try {
            // this is only discovering major database instances - and not all
            // services
            // hence, there's probably a 1-1 mapping between sid & db FIX THIS

            // try a ps to get all sid's
            log_.debug("Finding oracle processes"); //$NON-NLS-1$

            oracle_home = parent.getConfigFile().getRealFile();

            String psres = getPsCache();

            if (! isWindows_ && ! isOpenVms_ ){
                while (true) {
                    i = psres.indexOf("ora_pmon"); //$NON-NLS-1$
                    if (i > 0) {
                        psres = psres.substring(i + "ora_pmon_".length()).trim(); //$NON-NLS-1$

                        String[] res = psres.split("\\s"); //$NON-NLS-1$
                        if (res.length >= 1) {
                            oracle_sid = res[0];
                        } else {
                            break;
                        }
                        setOraHome(oracle_home);

                        log_.debug("findInstances: used ps to find sid:" + oracle_sid); //$NON-NLS-1$
                        // note that this is using the home discovered previously
                        add_instance(parent, oracle_sid.toLowerCase(), oracle_home, v);
                    } else {
                        break;
                    }
                }
            }
            if (isWindows_){
                String[] procs = psres.split("\n"); //$NON-NLS-1$
                HashMap locatedInstances = new HashMap();
                for (i=0; i < procs.length; i++){
                    log_.debug("Processing: " + procs[i]); //$NON-NLS-1$
                    Matcher m = Pattern.compile("oracle.exe\\s+(\\S+)",Pattern.CASE_INSENSITIVE).matcher(procs[i].toLowerCase()); //$NON-NLS-1$
                    if (m.find()){
                        oracle_sid = m.group(1);
                        //Check to ensure we have not already found this 
                        //instance
                        if (! locatedInstances.containsKey(oracle_sid)){
                            log_.debug("findInstances: found " + oracle_sid); //$NON-NLS-1$
                            locatedInstances.put(oracle_sid,oracle_sid);
                            setOraHome(oracle_home);
                            add_instance(parent,oracle_sid.toLowerCase(), oracle_home, v);
                        }
                        else{
                            log_.debug("findInstances: already found " + oracle_sid); //$NON-NLS-1$
                        }
                    }
                    else{
                        log_.debug("findInstances: could not find SID in " + procs[i]); //$NON-NLS-1$
                    }
                }
		//  While we are here, lets try to guess some sids based
                //  on registry key names
                String[] registry_sids = searchForSidsInRegistry();
                log_.debug("findInstances: checking sids from registry"); //$NON-NLS-1$
                for (i=0; i <registry_sids.length; i++){
                    oracle_sid = registry_sids[i].toLowerCase();
                    log_.debug("findInstances checking " + oracle_sid); //$NON-NLS-1$
                    if (! locatedInstances.containsKey(oracle_sid)){
                         log_.debug("findInstances: found " + oracle_sid + "in registry"); //$NON-NLS-1$ //$NON-NLS-2$
                         locatedInstances.put(oracle_sid,oracle_sid);
                         setOraHome(oracle_home);
                         add_instance(parent,oracle_sid.toLowerCase(), oracle_home, v);
            }	
                 }
	    }	
            if (isOpenVms_){
                String[] procs = psres.split("\n"); //$NON-NLS-1$
                HashMap locatedInstances = new HashMap();
                for (i=0; i < procs.length; i++){
                    log_.debug("Processing: " + procs[i]); //$NON-NLS-1$
                    Matcher m = Pattern.compile("ORA_(\\S+)_PMON").matcher(procs[i].toUpperCase()); //$NON-NLS-1$
                    if (m.find()){
                        oracle_sid = m.group(1);
                        //Check to ensure we have not already found this 
                        //instance
                        if (! locatedInstances.containsKey(oracle_sid)){
                            log_.debug("findInstances: found " + oracle_sid); //$NON-NLS-1$
                            locatedInstances.put(oracle_sid,oracle_sid);
                            setOraHome(oracle_home);
                            add_instance(parent,oracle_sid.toLowerCase(), oracle_home, v);
                        } else{
                            log_.debug("findInstances: already found " + oracle_sid); //$NON-NLS-1$
                        }
                    } else{
                        log_.debug("findInstances: could not find SID in " + procs[i]); //$NON-NLS-1$
                    }
                }
            }
            if (v.size() <= 0) {
                throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.5"); //$NON-NLS-1$
            }
            serverMap_.put(parent, v);
        } catch (Exception ex) {
            // either an IO or an authentication error, which is fatal
            log_.debug("findInstances: Exception finding instances", ex); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.6", new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleInstance[] oarray = new OracleInstance[v.size()];
        return (OracleInstance[]) v.toArray(oarray);
    }

    private void add_instance(OracleServer parent, String oracle_sid, String oracle_home, Vector v) throws AgentException, AttributeNotSetException,
            com.collation.platform.os.OsException {

        OracleInstance oinst = null;

        int oracle_port = getOraclePort2(oracle_sid);
        if (oracle_port < 0) {
            log_.info(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.I.0", new Object[]{oracle_sid})); //$NON-NLS-1$
            return;
        }

        String oracle_dbname = getOracleDbname(oracle_sid, oracle_port);

        log_.debug("[add_instance] Using oracle home as : " + oracle_home); //$NON-NLS-1$

        /*
         * Commented as oracle_home is already got in findInstances.
         * getOracleHome method will return only the oracle_base path, because
         * the condition core_dump_dest will return $ORACLE_BASE/admin/<SID>/cdump
         * and not the oracle_home path
         */
        // oracle_home = getOracleHome(oracle_sid, oracle_port, oracle_home);
        // prevent duplicate sids
        if (sidMap_.get(oracle_sid + oracle_home) != null) {
            return;
        }

        // prevent duplicates
        for (int j = 0; j < v.size(); j++) {
            oinst = (OracleInstance) v.get(j);
            if (oinst.getSID().compareTo(oracle_sid) == 0 && oinst.getHome().compareTo(oracle_home) == 0) {
                return;
            }
        }

        oinst = (OracleInstance) ModelFactory.newInstance(OracleInstance.class);
        log_.debug("findInstances: oracle sid: " + oracle_sid); //$NON-NLS-1$
        log_.debug("findInstances: oracle home: " + oracle_home); //$NON-NLS-1$
        log_.debug("findInstances: oracle port: " + oracle_port); //$NON-NLS-1$
        oinst.setParent(parent);
        oinst.setSID(oracle_sid);
        oinst.setName(oracle_sid);
        oinst.setKeyName(oracle_sid);
        oinst.setHome(oracle_home);
        oinst.setPort(oracle_port);
        oinst.setHost(os_.getComputerSystem());
        LogicalContent[] lcs = findInstanceConfigfiles(oracle_home, oracle_sid);
        if (lcs != null) {
            AppConfig[] acs = new AppConfig[lcs.length];
            for (int i = 0; i < lcs.length; i++) {
                acs[i] = (AppConfig) ModelFactory.newInstance(AppConfig.class);
                acs[i].setParent(oinst);
                acs[i].setContent(lcs[i]);
            }
            oinst.setConfigContents(acs);
        }

        BindAddress ba = (BindAddress) ModelFactory.newInstance(BindAddress.class);
        ba.setType(BindAddress.TCP_PORT);
        ba.setPortNumber(oracle_port);
        ba.setIpAddress(primaryIp_);
        ba.setPrimaryIpAddress(primaryIp_);
        oinst.setPrimarySAP(ba);

        v.add(oinst);
        instanceList_.add(oinst);
        sidMap_.put((oracle_sid + oracle_home), oracle_sid);
    }

    private String findInitDotOra(String[] dentrys, String regex){
	Pattern pattern = Pattern.compile(regex);
	//for (String pfile_entry : dentrys){
        String pfile_entry = null;
	for (int i =0; i < dentrys.length; i++){
            pfile_entry = dentrys[i];
	    log_.debug("findInitDotOra() Checking " + pfile_entry); //$NON-NLS-1$
	    Matcher m = pattern.matcher(pfile_entry.toLowerCase());
	    if (m.find()){
		return(pfile_entry);
	    }
	}
	return(null); 
    }

    private String cleanUpFileName(String file){
	String clean_file = file;
	log_.debug("cleanUpFileName: attempting to clean up " + file); //$NON-NLS-1$
	if (! isWindows_){
	    Pattern extraSlashes = Pattern.compile("/{2,}"); //$NON-NLS-1$
	    clean_file = extraSlashes.matcher(file).replaceAll("/"); //$NON-NLS-1$
	}
	if (isWindows_){
	    //Convert to backslashes
	    Pattern backSlashes = Pattern.compile("/"); //$NON-NLS-1$
	    //this is real screwed up.  Because a backslash is an
	    //escape character you have to escape it. That gives you a
	    //String that is a backslash, but a backslash is also a regex
	    //escape so you need to escape it with a backslash. 
	    //which nets four backslashes to get one
	    //see http://www.regular-expressions.info/java.html
	    clean_file = backSlashes.matcher(file).replaceAll("\\\\"); //$NON-NLS-1$
	    Pattern extraBackSlashes = Pattern.compile("\\\\{2,}"); //$NON-NLS-1$
	    clean_file = extraBackSlashes.matcher(clean_file).replaceAll("\\\\"); //$NON-NLS-1$
	}
	if (file.compareToIgnoreCase(clean_file) != 0){
		log_.debug("cleanUpFileName: changed: " + file + " to " + clean_file); //$NON-NLS-1$ //$NON-NLS-2$
	}
	return(clean_file);
    } 

    private LogicalContent[] findInstanceConfigfiles(String oracle_home, String oracle_sid) {
        log_.debug("Getting Logical Contents"); //$NON-NLS-1$

        ArrayList lcs = new ArrayList();
        String initora = oracle_home + "/admin/" + oracle_sid + "/pfile/" + "init*.ora"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        String initora2 = oracle_home + "/admin/" + oracle_sid + "/pfile/" + "init.ora.[0-9]*"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        String file = null;

        String[] files = { null, // will be init.ora
                oracle_home + "/network/admin/" + "sqlnet.ora", oracle_home + "/network/admin/" + "tnsnames.ora", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
                oracle_home + "/network/admin/" + "listener.ora" }; //$NON-NLS-1$ //$NON-NLS-2$

        String[] names = { "init.ora", "sqlnet.ora", "tnsnames.ora", "listener.ora" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$


        // OpenVms commands and directories are a bit different, so we need to adjust the values.
        String initorapath = "/" + oracle_home.substring(0,oracle_home.length()-1) + ".admin." + oracle_sid + ".pfile]"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        String spOpenVmsPath = "/" + oracle_home.substring(0,oracle_home.length()-1) + ".dbs]"; //$NON-NLS-1$ //$NON-NLS-2$
        String[] openVmsFiles = { null, // will be init.ora
                "/" + oracle_home.substring(0, oracle_home.length()-1) + ".network.admin]" + "sqlnet.ora",  //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                "/" + oracle_home.substring(0, oracle_home.length()-1) + ".network.admin]" + "tnsnames.ora", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                "/" + oracle_home.substring(0, oracle_home.length()-1) + ".network.admin]" + "listener.ora" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

        // Since we don't know what the name format of the init.ora file is
		// exactally, we need to search for it.
		String pfile_dir = oracle_home + "/admin/" + oracle_sid + "/pfile/"; //$NON-NLS-1$ //$NON-NLS-2$
		if(isOpenVms_) {
			pfile_dir = initorapath;
        	files = openVmsFiles;
        }
        try {
			String[] pfile_dir_list = os_.listDirectory(pfile_dir, false, null);
			// First try init*.ora
			log_.debug("findInstanceConfigfile: looking for init*.ora"); //$NON-NLS-1$
			if(!isOpenVms_) {
				file = findInitDotOra(pfile_dir_list,"(init[^\\.]\\.ora)"); //$NON-NLS-1$
            } else {
				file = findInitDotOra(pfile_dir_list,"(init[^\\.]*\\.ora;\\S)"); //$NON-NLS-1$
            }
			if (file == null) {
				// ok, that didnt work, try init.ora.[0-9]*
				log_.debug("findInstanceConfigfile: looking for init.ora.[0-9]*"); //$NON-NLS-1$
				if(!isOpenVms_) {
					file = findInitDotOra(pfile_dir_list,"(init\\.ora\\.[0-9]\\S+)"); //$NON-NLS-1$
				} else {
					file = findInitDotOra(pfile_dir_list,"(init\\.ora\\.[0-9]\\S+;\\S)"); //$NON-NLS-1$
				}
			}
			// If we found something, lets prepend the path
			if (file != null) {
				file = pfile_dir + file;
			}
        } catch (Exception ex) {
			log_.debug("findInstanceConfigfile: Os.listDirectory failed for: " //$NON-NLS-1$
					+ pfile_dir + ", exception:" + ex.getMessage()); //$NON-NLS-1$
        }

		if (file == null) {
			// OK, didnt find init.ora. Lets try an SPFILE
			String spfile_dir = oracle_home + "/dbs/"; //$NON-NLS-1$
			if(isOpenVms_) {
				spfile_dir = spOpenVmsPath;
			}
            try {
				String[] spfile_dir_list = os_.listDirectory(spfile_dir, false, null);
				log_.debug("findInstanceConfigfile: looking for spfile"	+ oracle_sid + ".ora"); //$NON-NLS-1$ //$NON-NLS-2$
				file = findInitDotOra(spfile_dir_list, "spfile"	+ oracle_sid.toLowerCase() + "\\.ora"); //$NON-NLS-1$ //$NON-NLS-2$
				if (file != null) {
					file = spfile_dir + file;
                }
            } catch (Exception ex) {
				log_.debug("findInstanceConfigfile: Os.listDirectory failed for: " //$NON-NLS-1$
								+ spfile_dir + ", exception:" + ex.getMessage()); //$NON-NLS-1$
            }
        }

		int first_file = 1;
        if (file != null && file.length() > 0) {
            files[0] = file;
			first_file = 0;
        }

		// Here if first_file is 1 then we couldnt find the init
		// file so we will skip it and try to grab the others.

		String[] paths = new String[files.length - first_file];
		OsException[] oses = new OsException[files.length - first_file];
		int[] maxSizes = new int[files.length - first_file];
		boolean[] recursives = new boolean[files.length - first_file];

		for (int i = 0; i < paths.length; i++) {
			paths[i] = cleanUpFileName(files[i + first_file]);
			oses[i] = null;
			maxSizes[i] = os_.ALL_CONTENT;
			recursives[i] = false;
                }

		LogicalContent[] instance_configs = os_.getFiles(paths, oses, maxSizes,
				recursives);

		for (int i = 0; i < instance_configs.length; i++) {
			if ((oses[i] == null) && (instance_configs[i] != null)) {
				if (!isOpenVms_) {
					instance_configs[i].setFixedPath(paths[i]);
				} else {
					// Remove leading '/' for OpenVms
					instance_configs[i].setFixedPath(paths[i].substring(0,paths[i].length()-1));
            }
				log_.debug("findInstanceConfigs: adding content for" + paths[i]); //$NON-NLS-1$
				lcs.add(instance_configs[i]);
			} else {
				log_.debug("findInstanceConfigs: Could not get content for: " + paths[i]); //$NON-NLS-1$
            }
		}

        LogicalContent[] results = (LogicalContent[]) lcs.toArray(new LogicalContent[lcs.size()]);
        return results;
    }

    /**
     * Find the listeners registered in each oracle home using ps & lsof. This
     * is used when the config files are not readable. Allows for multiple
     * listeners/server.
     * 
     * If no ports are found, no exception is thrown.
     */
    public OracleListener[] findListener2(OracleServer parent) throws AgentException, InterruptedException {
        OracleListener listener = null;
        Vector lv = new Vector(5);

        if (portList == null) {
            portList = getPortList();
        }

        String home = null;
        try {
            home = parent.getConfigFile().getRealFile();
        } catch (Exception ex) {
            log_.debug("findListener2: Unable to get Oracle home:" + ex); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.7", new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex); //$NON-NLS-1$
        }

        // iterate over all listeners found via ps
        for (Iterator iter = portList.keySet().iterator(); iter.hasNext();) {
            String l = (String) iter.next();
            if (isInDirectory(home, l)) {

                // found a match

                String str = l.trim();

                listener = (OracleListener) ModelFactory.newInstance(OracleListener.class);
                listener.setParent(parent);
                listener.setName(str);
                lv.add(listener);

                // now convert ports to BindAddresses
                ArrayList ports = (ArrayList) portList.get(l);
                if (ports == null) {
                    log_.warning(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.W.0", new Object[]{l})); //$NON-NLS-1$
                    continue;
                }

                BindAddress[] oarray = new BindAddress[ports.size()];
                for (int i = 0; i < ports.size(); i++) {
                    BindAddress ba = (BindAddress) ports.get(i);
                    oarray[i] = ba;
                }
                listener.setBindAddresses(oarray);
            }
        } // for each listener process

        // convert the vector to an array
        log_.debug("findListener2, found: " + lv.size() + " listeners"); //$NON-NLS-1$ //$NON-NLS-2$
        OracleListener[] oarray = new OracleListener[lv.size()];
        return (OracleListener[]) lv.toArray(oarray);
    } // findListener

    /**
     * Find all databases in a "Oracle Server" (a computer). One db may map to
     * many instances. Databases are found by going through the instance list
     * for unique db's. [todo: the code currently assumes a 1-1 mapping FIX
     * THIS]
     * 
     * This is the function which does the majority of Oracle discovery.
     * 
     */
    public OracleDatabase[] findDatabases(OracleServer parent) throws AgentException {
        OracleInstance[] oinstv;
        OracleInstance oinst;
        String res;
        Connection conn = null;
        Vector v = new Vector(5);
        OracleDatabase odb;
        int connCount = 0;

        try {
            // disables to break up results
            // oinstv = parent.getInstances();
            Vector ov = (Vector) serverMap_.get(parent);
            OracleInstance[] iarray = new OracleInstance[ov.size()];
            oinstv = (OracleInstance[]) ov.toArray(iarray);

            // Find all matches
            for (int i = 0; i < oinstv.length; i++) {
                oinst = oinstv[i];

                conn = getConnection(oinst.getPort(), null, oinst.getSID());

                //
                // This code relies upon the instance being up to discover
                // anything about the database. Alternatively, it's possible
                // to scan through init.ora to get some info. This code will
                // just leave everything as null when an instance can't be
                // connected to using the port/uid/pwd.
                //
                if (conn == null) {
                    oinst.setStatus(OracleInstance.STOPPED);
                } else {
                    try {
                        connCount++;
                        odb = (OracleDatabase) ModelFactory.newInstance(OracleDatabase.class);
                        odb.setParent(oinst);
                        odb.setDBVersion(getDBVersionInfo(conn));
                        odb.setDBName(getDBProductInfo(conn));
                        odb.setName(getGlobalName(conn));

                        odb.setTableSpaces(findTableSpaces(odb, conn));
                        odb.setInitValues(findInitValues(odb, conn));
                        // disable set schema for better performance
                        odb.setSchemas(findSchema(odb, conn));
                        OracleSchema[] schemas = odb.getSchemas();
                        if (schemas != null && schemas.length > 0) {
                            ArrayList modules = new ArrayList();
                            for (int s = 0; s < schemas.length; s++) {
                                try {
                                    OracleModule m = (OracleModule) ModelFactory.newInstance(OracleModule.class);
                                    m.setParent(oinst);
                                    m.setSchema(schemas[s]);
                                    m.setFileName(schemas[s].getOwner());
                                    m.setName(schemas[s].getOwner());
                                    modules.add(m);
                                } catch (Exception ex) {
                                    log_.debug("Unexpected exception", ex); //$NON-NLS-1$
                                }
                            }
                            oinst.setModules((SoftwareModule[]) modules.toArray(new SoftwareModule[0]));
                        }
                        odb.setSchemasRawData(findRawSchema(odb, oinst, conn));
                        odb.setDataFiles(findDataFiles(odb, conn));
                        odb.setControlFiles(findControlFiles(odb, conn));
                        odb.setRedoLogFiles(findRedoLogs(odb, conn));
                        // odb.setSegments(findSegments(odb, conn)); FIX THIS

                        // DB Link - Extended Feature
                        if (isSensorExtended()) {
                            log_.debug("Extended sensor is enabled"); //$NON-NLS-1$
                            odb.setDBLinks(findDBLinks(odb, conn, oinst.getSID()));
                            log_.debug("Discovering Metadata info..."); //$NON-NLS-1$
                            odb.setSchemaContent(findSchemaObjects(odb, oinst, conn));
                        }

                        // instance info is set here as well
                        OracleBackgroundProcess[] bg = findBGProcesses(oinst, conn);
                        oinst.setBackgroundProcesses(bg);

                        OracleServerProcess[] servers = findServerProcesses(oinst, conn);
                        oinst.setServerProcesses(servers);

                        ProcessPool pool = (ProcessPool) ModelFactory.newInstance(ProcessPool.class);
                        pool = (ProcessPool) ModelFactory.newInstance(ProcessPool.class);
                        pool.setParent(oinst);
                        pool.setName("process pool"); //$NON-NLS-1$
                        pool.setRuntimeProcesses(servers);

                        oinst.setProcessPools(new ProcessPool[] { pool });

                        oinst.setSGAValues(findSGAValues(oinst, conn));
                        oinst.setDatabase(odb);
                        oinst.setStatus(OracleInstance.RUNNING);

                        v.add(odb);
                    } catch (AgentException ex) {
                        log_.debug("findDatabases, continuing from general exception:", ex); //$NON-NLS-1$
                    }
                }

            }
        } catch (Exception ex) {
            log_.debug("findDatabases, general exception:", ex); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.8", new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex); // let //$NON-NLS-1$
            // the
            // parents
            // worry
            // about
            // it
        }

        if (connCount == 0) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.9")); //$NON-NLS-1$
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.9"); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleDatabase[] oarray = new OracleDatabase[v.size()];
        return (OracleDatabase[]) v.toArray(oarray);
    }

    /**
     * Establishes and returns a JDBC Oracle connection, using the given port &
     * SID. Gets the uid/pwd from the seed. Returns null on error (and logs it.)
     * Note that this is using auto-commit in read-committed mode.
     * 
     * The format of the Oracle thin jdbc driver url is:
     * jdbc:oracle:thin:@machineName:port:SID
     */
    private Connection getConnection(int portNum, String addr, String sid) throws AgentException {
        Connection conn = null;
        String connStr;
        DatabaseMetaData meta;
        String version = null;
        String str = null;

        // first check for suppressed sids/ports
        if (suppressSIDs_.containsKey(sid.toLowerCase())){
            log_.debug("Found SID on supression list: " + sid); //$NON-NLS-1$
            return null;
        }
        if (suppressPorts_.containsKey(new Integer(portNum))){
            log_.debug("Found port on supression list: " + portNum); //$NON-NLS-1$
            return null;
        }
        // we may need connections many times
        conn = (Connection) connPool.get(sid + portNum);
        if (conn != null)
            return conn;

        if (addr == null) {
            addr = ipName_;
        }

        connStr = "jdbc:oracle:thin:@" + addr + ":" + portNum + ":" + sid; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

        if (portNum <= 0) {
            log_.error(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.E.10", new Object[]{connStr})); //$NON-NLS-1$
            return null;
        }

        if (!testPort(addr, portNum)) {
            log_.debug("Address unreachable:" + addr + ":" + portNum); //$NON-NLS-1$ //$NON-NLS-2$
            return null;
        }

        try {
            if (authList_ == null){
                authList_ = AuthManager.getAuth(AuthManager.ORACLEAUTH, primaryIp_);
            }
        } catch (Exception ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.11"), ex); //$NON-NLS-1$
        }
        if (authList_ == null) {
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.12"); //$NON-NLS-1$
        }
        if (authList_ != null) {
            for (Iterator it = authList_.iterator(); it.hasNext() && conn == null;) {
                // try each uid/pwd in the list until one works
                DbAuth auth = (DbAuth) it.next();
                String oraUID = auth.getUserName();
                String oraPWD = auth.getPassword();

                try {

                    log_.debug("getConnection: about to connect to: " + connStr + "/" + auth.getUserName()); //$NON-NLS-1$ //$NON-NLS-2$

                    if (driver_ == null){
                        //avoid calling DriverManager methods, they are
                         //class serialized
                        driver_ = DriverManager.getDriver(connStr);
                    }
                    Properties authProps = new Properties();
                    authProps.setProperty("user",oraUID);  //$NON-NLS-1$
                    authProps.setProperty("password",oraPWD); //$NON-NLS-1$
                    if (useConnectorThreads_){
                    
                        OracleJDBCConnector connThread = new OracleJDBCConnector(driver_,connStr,authProps,log_);
                        log_.debug("getConnection: about to start JDBCConnector");  //$NON-NLS-1$
                        synchronized (connThread) {
                            connThread.start();
			    log_.debug("getConnection: waiting for JDBCConnector"); //$NON-NLS-1$
			    try{
                                connThread.wait(connThreadTimeout_);
                                log_.debug("OracleJDBCConnector.wait() completed"); //$NON-NLS-1$
                            }
                            catch (InterruptedException iex){
                            } 
                        }
                        conn = connThread.getConnection();
                        if (conn == null){
                             log_.debug("getConnection: no connection created.");
                             Exception cex = connThread.getException();
                             if (cex != null){
                                 log_.debug("getConnection: unable to connect w/conn string:" + connStr + ", ex: " + cex); //$NON-NLS-1$ //$NON-NLS-2$
                             }
                             continue;
                        }
                    }
                         
                    else{
                        conn = driver_.connect(connStr,authProps);
                    }
                    log_.debug("getConnection: got connection to " + connStr);  //$NON-NLS-1$
                    meta = conn.getMetaData();
                    log_.debug("getConnection: got meta data");  //$NON-NLS-1$ 

                    str = meta.getDatabaseProductVersion();
                } catch (Exception ex) {
                    log_.debug("getConnection: unable to connect w/conn string:" + connStr + ", ex: " + ex); //$NON-NLS-1$ //$NON-NLS-2$
                    conn = null;
                }

            }
        }

        if (conn == null) {
            // an expected error if the instance is down
            log_.info(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.I.1", new Object[]{connStr})); //$NON-NLS-1$
            // log_.debug("getConnection: debug info, current classpath: " +
            // System.getProperty("java.class.path"));
        }

        if (conn != null) {
            try {
                //
                // look for the number X in X.Y.Z in the version string
                //
                Pattern pattern = Pattern.compile("(.*?)(\\d*?)\\.(\\d*?)\\.(\\d*?)", Pattern.MULTILINE); //$NON-NLS-1$
                Matcher matcher = pattern.matcher(str);

                // Find first numerical match
                if (matcher.find()) {
                    version = matcher.group(2).trim();
                }
                oracleVersion_ = new Integer(version).intValue();

                log_.info(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.I.2", new Object[]{sid, hostName_, version})); //$NON-NLS-1$

            } catch (Exception ex) {
                // this could happen in the group(2) if there's no match
                log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.1"), ex); //$NON-NLS-1$
            }

        }

        if (conn != null && (oracleVersion_ != 8 && oracleVersion_ != 9 && oracleVersion_ != 10)) {
            log_.warning(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.W.2", new Object[]{str})); //$NON-NLS-1$
        }

        connPool.put((sid + portNum), conn);

        return conn;
    }

    /**
     * Gets Oracle product description from the given connection. Returns null
     * on error (and logs it.)
     */
    private String getDBVersionInfo(Connection conn) {
        String str = null;
        String version = null;
        String query = "select banner from v$version"; //$NON-NLS-1$

        try {
            Statement stmt;
            ResultSet rs;

            stmt = conn.createStatement();
            log_.debug("getDBVersionInfo: about to execute:" + query); //$NON-NLS-1$
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                if (str.toLowerCase().startsWith("oracle")) { //$NON-NLS-1$
                    break;
                }
            }
            stmt.close();

        } catch (SQLException ex) {
            // this should never happen, but might as well try to continue
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.3"), ex); //$NON-NLS-1$
        }

        if (str != null) {
            str = str.substring(str.indexOf("Release")).trim(); //$NON-NLS-1$
        }
        log_.debug("getDBVersionInfo: oracle version:" + str); //$NON-NLS-1$
        return str;
    }

    /**
     * Gets Oracle version description from the given connection. Returns null
     * on error (and logs it.)
     */
    private String getDBProductInfo(Connection conn) {
        String str = null;
        String version = null;
        String query = "select banner from v$version"; //$NON-NLS-1$

        try {
            Statement stmt;
            ResultSet rs;

            stmt = conn.createStatement();
            log_.debug("getDBProductInfo: about to execute:" + query); //$NON-NLS-1$
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                if (str.toLowerCase().startsWith("oracle")) { //$NON-NLS-1$
                    break;
                }
            }
            stmt.close();

        } catch (SQLException ex) {
            // this should never happen, but might as well try to continue
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.4"), ex); //$NON-NLS-1$
        }

        if (str != null) {
            str = str.substring(0, str.indexOf("Release")).trim(); //$NON-NLS-1$
        }
        log_.debug("getDBProductInfo: oracle product name:" + str); //$NON-NLS-1$
        return str;
    }

    /**
     * Gets global database name from given connection. Returns null on error
     * (and logs it.)
     */
    private String getGlobalName(Connection conn) throws AgentException {
        String str = null;
        String version = null;
        String query = "select global_name from global_name"; //$NON-NLS-1$

        try {
            Statement stmt;
            ResultSet rs;

            stmt = conn.createStatement();
            log_.debug("getGlobalName: about to execute:" + query); //$NON-NLS-1$
            rs = stmt.executeQuery(query);

            rs.next();
            str = rs.getString(1);
            log_.debug("getGlobalName: oracle global database name:" + str); //$NON-NLS-1$
            stmt.close();

        } catch (SQLException ex) {
            // this should never happen, and is fatal as the Global Name is a
            // key
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.13", new Object[]{InterMessageUtils.getIntactMessage(ex)}, ex); //$NON-NLS-1$
        }

        return str;
    }

    /**
     * Tries to find a connection to the given sid on the list of ports found in
     * the seed. Returns the port which worked and adds the connection to pool.
     * 
     * Extracts Oracle listening port from ps/lsof data. assumes sid parameter
     * is lower cased
     * 
     * If the port can't be obtained, -1 is returned.
     */
    private int getOraclePort2(String sid) throws AgentException {
        if (portList == null) {
            portList = getPortList();
        }

        try {
            HashMap portMap = new HashMap();
            for (Iterator iter = portList.keySet().iterator(); iter.hasNext();) {
                String l = (String) iter.next();

                // try this listener
                ArrayList ports = (ArrayList) portList.get(l);
                if (ports == null) {
                    continue;
                }

                for (int i = 0; i < ports.size(); i++) {
                    BindAddress ba = (BindAddress) ports.get(i);
                    String[] addrs = getAddrs(ba);
                    int port = ba.getPortNumber();
                    Integer p = new Integer(port);
                    // try all listening interfaces
                    for (int j = 0; j < addrs.length; j++) {
                        String addr = addrs[j];
                        if (portMap.get(addr + ":" + p) != null) { //$NON-NLS-1$
                            continue;
                        }

                        /**
                         * Restrict discovery only to those ports which are part
                         * of the original seed. Currently, separate oracle
                         * seeds are created for each listener - and it makes
                         * sense that each agent should only go through a single
                         * listener, rather than duplicate all the work.
                         */
                        boolean foundIt = false;
                        for (int k = 0; k < seedPorts_.length; k++) {
                            if (seedPorts_[k] == port) {
                                foundIt = true;
                                break;
                            }
                        }
                        if (!foundIt) {
                            continue;
                        }

                        portMap.put(addr + ":" + p, p); //$NON-NLS-1$

                        Connection conn = getConnection(port, addr, sid);
                        if (conn != null) {
                            log_.debug("getOraclePort2: " + port); //$NON-NLS-1$
                            return port;
                        }
                    }
                }
            }
        } catch (Exception ex) {
            log_.debug("unable to get listening port:" + ex); //$NON-NLS-1$
        }

        return -1;
    }

    /**
     * Get oracle db name straight from a SELECT global_name. If the name can't
     * be found, returns null.
     */
    private String getOracleDbname(String sid, int port) throws AgentException {
        Connection conn = null;

        try {
            conn = getConnection(port, null, sid);
            if (conn != null) {
                return getGlobalName(conn);
            }
        } catch (Exception ex) {
            log_.debug("unable to get global database name:" + ex); //$NON-NLS-1$
        }

        return null;
    }

    /**
     * Get oracle home straight from a SELECT stmt. If the name can't be found,
     * returns null.
     */
    private String getOracleHome(String sid, int port, String oldHome) throws AgentException {
        Connection conn = null;
        String newHome = null;

        conn = getConnection(port, null, sid);
        if (conn != null) {

            ResultSet rs = null;
            PreparedStatement ps = null;

            String query = "select value from v$parameter where name = 'core_dump_dest'"; //$NON-NLS-1$

            try {
                log_.debug("getOracleHome about to prepare: " + query); //$NON-NLS-1$
                ps = conn.prepareStatement(query);
                rs = ps.executeQuery();
                if (rs.next()) {
                    String str = rs.getString(1);
                    int i = str.lastIndexOf("admin"); //$NON-NLS-1$
                    if (i > 0) {
                        newHome = str.substring(0, i - 1);
                    }
                }

                ps.close();

            } catch (SQLException ex) {
                // this may be due to insufficient privileges to see the view
                // v$parameters requires SELECT_CATALOG_ROLE
                log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.14")); //$NON-NLS-1$
            }
        }

        if (newHome == null) {
            log_.warning(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.W.5", new Object[]{oldHome})); //$NON-NLS-1$
            newHome = oldHome;
        }
        log_.debug("getOracleHome using: " + newHome); //$NON-NLS-1$
        return newHome;
    }

    /**
     * Finds all table spaces visible to the connection. Exception thrown if no
     * spaces found.
     */
    public OracleTableSpace[] findTableSpaces(OracleDatabase parent, Connection conn) throws AgentException {
        OracleTableSpace ots;
        Vector v = new Vector(10);
        String str;
        long i;

        // can this query may return multiple copies of the same tablespace?
        // check this out, when extending a tspace across > 1 file
        // if so, then the size should be the sum of all entries
        // FIX THIS
        String query = "select tablespace_name, bytes " + " from dba_data_files order by tablespace_name"; //$NON-NLS-1$ //$NON-NLS-2$
        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                i = rs.getLong(2);
                log_.debug("findTableSpaces, space: " + str + ", size: " + i); //$NON-NLS-1$ //$NON-NLS-2$

                ots = (OracleTableSpace) ModelFactory.newInstance(OracleTableSpace.class);
                ots.setParent(parent);
                ots.setName(str);
                ots.setSize(i);

                v.add(ots);
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.15"), ex); //$NON-NLS-1$
            log_.info(DiscoverAgentLocalizedMessages.getString("OracleAgent.I.3")); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.6")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleTableSpace[] oarray = new OracleTableSpace[v.size()];
        return (OracleTableSpace[]) v.toArray(oarray);
    }

    /**
     * Finds all redo logs visible to the connection. Exception thrown if SQL
     * statement fails.
     */
    public OracleRedoLogFile[] findRedoLogs(OracleDatabase parent, Connection conn) throws AgentException {
        OracleRedoLogFile orf;
        Vector v = new Vector(10);
        String str;
        long i;

        String query = "select y.member, x.bytes from " + "v$log x, v$logfile y where " + "x.group# = y.group# order by x.group#"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                i = rs.getLong(2);
                log_.debug("findRedoLogs, space: " + str + ", size: " + i); //$NON-NLS-1$ //$NON-NLS-2$

                orf = (OracleRedoLogFile) ModelFactory.newInstance(OracleRedoLogFile.class);
                orf.setParent(parent);
                orf.setName(str);
                orf.setSize(i);

                v.add(orf);
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.16"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.7")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleRedoLogFile[] oarray = new OracleRedoLogFile[v.size()];
        return (OracleRedoLogFile[]) v.toArray(oarray);
    }

    // Oracle9 list of initialization parameters
    static final String names[] = { "db_block_buffers", "db_cache_size", "db_domain", "db_files", "db_name", "dml_locks", "instance_name", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
            "java_pool_size", "large_pool_size", "log_buffer", "max_rollback_segments", "open_cursors", "pga_aggregate_target", "processes", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$
            "service_names", "sessions", "sga_max_size", "shared_pool_size", "transactions" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$

    /**
     * Finds all initialization values pertinent to the given version of Oracle.
     * Also gets the description of each initialization value.
     */
    public OracleInitValue[] findInitValues(OracleDatabase parent, Connection conn) throws AgentException {
        OracleInitValue oval;
        Vector v = new Vector(10);
        ResultSet rs = null;
        Statement stmt = null;
        String name = ""; //$NON-NLS-1$
        String value = ""; //$NON-NLS-1$
        String desc = ""; //$NON-NLS-1$

        String query = "select name, value, description from v$parameter where upper(ISSYS_MODIFIABLE) != 'IMMEDIATE'"; //$NON-NLS-1$

        try {
            log_.debug("findInitValues about to prepare: " + query); //$NON-NLS-1$
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                name = rs.getString(1);
                value = rs.getString(2);
                desc = rs.getString(3);
                log_.debug("findInitValues found: " + name + " = " + value + " [" + desc + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

                oval = (OracleInitValue) ModelFactory.newInstance(OracleInitValue.class);
                oval.setParent(parent);
                oval.setName(name);
                oval.setValue(value);
                oval.setDescription(desc);

                v.add(oval);
            }

            stmt.close();
        } catch (SQLException ex) {
            // this may be due to insufficient privileges to see the view
            // v$parameters requires SELECT_CATALOG_ROLE
            log_.warning(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.W.8", new Object[]{name}), ex); //$NON-NLS-1$
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.9")); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.10")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleInitValue[] oarray = new OracleInitValue[v.size()];

        return (OracleInitValue[]) v.toArray(oarray);
    }

    /**
     * Find all non system tables on the given connection. A system table is
     * owned by any users which are created by Oracle at database creation,
     * except for demo users and their tables. It's ok for no tables to be
     * found, although it's likely that there's a problem.
     */
    public OracleSchema[] findSchema(OracleDatabase parent, Connection conn) throws AgentException {
        OracleSchema os;
        OracleTableSpace ts;
        Vector v = new Vector(100);
        String str;
        String str2;
        String str3;

        // select all non-system tables, using a typical Oracle 9
        // system id list
        //
        /*
         * String query9 = "select owner, table_name, tablespace_name " + "from
         * sys.dba_tables " + "where owner not in
         * ('SYSTEM','SYS','WMSYS','CTXSYS','MDSYS'," +
         * "'ORDSYS','XDB','OUTLN','AURORA$JIS$UTILITY$','OSE$HTTP$ADMIN') " +
         * "order by owner, tablespace_name, table_name";
         */
        String query9 = "select distinct owner " + "from sys.dba_tables " + "where owner not in ('SYSTEM','SYS','WMSYS','CTXSYS','MDSYS'," //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                + "'ORDSYS','XDB','OUTLN','AURORA$JIS$UTILITY$','OSE$HTTP$ADMIN','DBSNMP','DMSYS','EXFSYS','SYSMAN') "  //$NON-NLS-1$
	            + "order by owner"; //$NON-NLS-1$

        log_.debug(query9);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query9);

            while (rs.next()) {
                str = rs.getString(1);
                // str2 = rs.getString(2);
                // str3 = rs.getString(3);
                // log_.debug("findSchema, table: " + str + "." + str2 + ",
                // tablespace: " + str3);
                log_.debug("findSchema, owner: " + str); //$NON-NLS-1$

                os = (OracleSchema) ModelFactory.newInstance(OracleSchema.class);
                os.setParent(parent);
                os.setOwner(str);
                os.setName(str);
                /*
                 * ts = findSpaceByName(parent, str3); if (ts != null) {
                 * os.setTableSpace(ts); }
                 */
                v.add(os);

            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.17"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.11")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleSchema[] oarray = new OracleSchema[v.size()];
        return (OracleSchema[]) v.toArray(oarray);
    }

    /**
     * Raw/stripped version of findSchema.
     * 
     * Find all non system tables on the given connection. A system table is
     * owned by any users which are created by Oracle at database creation,
     * except for demo users and their tables. It's ok for no tables to be
     * found, although it's likely that there's a problem.
     */
    public LogicalContent findRawSchema(OracleDatabase parent, OracleInstance oinst, Connection conn) throws AgentException {
        StringBuffer buf = new StringBuffer(2000);
        String str;
        String str2;
        long count = 0;

        String query = "select count(*) " + "from sys.dba_tables " + "where owner not in ('SYSTEM','SYS','WMSYS','CTXSYS','MDSYS'," //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                + "'ORDSYS','XDB','OUTLN','AURORA$JIS$UTILITY$','OSE$HTTP$ADMIN','DBSNMP','DMSYS','EXFSYS','SYSMAN') "  //$NON-NLS-1$
                + "order by owner, table_name"; //$NON-NLS-1$

        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                count = rs.getLong(1);
                log_.debug("findRawSchema, count(*):" + count); //$NON-NLS-1$
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.18"), ex); //$NON-NLS-1$
        }

        if (count > TABLE_MAX) {
            buf.append(DiscoverAgentLocalizedMessages.getFormattedUIString("OracleAgent.R.0", new Object[]{new Long(count), new Integer(TABLE_MAX), new Integer(TABLE_MAX)}, Locale.ENGLISH)); //$NON-NLS-1$
        } else {

            buf.append(DiscoverAgentLocalizedMessages.getFormattedUIString("OracleAgent.R.1", new Object[]{new Long(count)}, Locale.ENGLISH)); //$NON-NLS-1$
        }

        // select all non-system tables, using a typical Oracle 9
        // system id list
        //
        query = "select owner, table_name " + "from sys.dba_tables " + "where owner not in ('SYSTEM','SYS','WMSYS','CTXSYS','MDSYS'," //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                + "'ORDSYS','XDB','OUTLN','AURORA$JIS$UTILITY$','OSE$HTTP$ADMIN','DBSNMP','DMSYS','EXFSYS','SYSMAN') "  //$NON-NLS-1$
                + "order by owner, table_name"; //$NON-NLS-1$
    
        int i=0;
        log_.debug(query);
        log_.debug("Table_Max: " + TABLE_MAX); //$NON-NLS-1$
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            for (i = 0; i < TABLE_MAX && rs.next(); i++) {
                str = rs.getString(1);
                str2 = rs.getString(2);
                // log_.debug("findRawSchema, table: " + str + "." + str2);

                buf.append(str);
                buf.append("."); //$NON-NLS-1$
                buf.append(str2);
                buf.append("\n"); //$NON-NLS-1$
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.19"), ex); //$NON-NLS-1$
        }

        // convert the buffer to a string
        // since we are allowing configuration of TABLE_MAX now
        // it's possible that 32764 buffer may not be large
        // enough.  We'll allow it to be 32764 or i*35, whichever
        // is greater
        int max_buffer = 32764;
        if ( (i * 35) > 32764 ) {
            max_buffer = i * 35;
        }

        log_.debug("maximum buffer length allowed: " + max_buffer); //$NON-NLS-1$
        log_.debug("actual buffer length: " + buf.length()); //$NON-NLS-1$
        str = null;

        if (buf.length() > max_buffer) {
            log_.warning("Oracle table information exceeded buffer length (" + max_buffer + ") with a length of (" + buf.length() + ")."); //$NON-NLS-1$
            log_.warning("Oracle table information will be truncated."); //$NON-NLS-1$
            str = buf.substring(0, max_buffer - 1);
        } else {
            str = buf.toString();
        }

        LogicalContent pkgContent = null;
        try {
            pkgContent = (LogicalContent) ModelFactory.newInstance(LogicalContent.class);
            pkgContent.setContent(str);
            pkgContent.setChecksum(SecurityUtils.getMessageDigest(str, "MD5")); //$NON-NLS-1$

            pkgContent.setURI(URIUtils.newURI(com.collation.platform.model.topology.app.db.oracle.OracleDatabase.class,
                    primaryIp_.getDotNotation() + oinst.getHome() + "/" + oinst.getSID(), "schemasRawData").toString()); //$NON-NLS-1$ //$NON-NLS-2$
        } catch (com.collation.platform.model.AttributeNotSetException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.19"), ex); //$NON-NLS-1$
        }

        return pkgContent;
    }

    /**
     * Find all currently running Oracle processes which are part of a
     * pre-defined set of possible Oracle background tasks. Fetches the name,
     * system pid and a brief description (defined in the v$bgprocess view.)
     */
    public OracleBackgroundProcess[] findBGProcesses(OracleInstance parent, Connection conn) throws AgentException {
        OracleBackgroundProcess bp;
        Vector v = new Vector(100);
        String str;
        String str2;
        int i = 0;

        String query = "select b.name, b.description, p.spid " + "from v$bgprocess b, v$process p where b.paddr = p.addr"; //$NON-NLS-1$ //$NON-NLS-2$

        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                str2 = rs.getString(2);
                if(!isOpenVms_) {
                    i = rs.getInt(3);
                } else {
                	// In OpenVms a pid is hex number
                	String hexpid = rs.getString(3);
                	if(hexpid != null && hexpid.length() > 0) {
                	   i = Integer.parseInt(hexpid,16);
                	}
                }
                log_.debug("findBGProcesses, process: " + str + ", pid: " + i + " [" + str2 + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

                bp = (OracleBackgroundProcess) ModelFactory.newInstance(OracleBackgroundProcess.class);
                bp.setParent(parent);
                bp.setName(str);
                bp.setDescription(str2);
                bp.setPid(i);
                v.add(bp);
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.20"), ex); //$NON-NLS-1$
        }

        query = "select b.name, p.spid " + "from v$dispatcher b, v$process p where b.paddr = p.addr"; //$NON-NLS-1$ //$NON-NLS-2$

        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                str2 = "dispatcher"; //$NON-NLS-1$
                if(!isOpenVms_) {
                    i = rs.getInt(2);
                } else {
                	// In OpenVms a pid is hex number
                	String hexpid = rs.getString(2);
                	if(hexpid != null && hexpid.length() > 0) {
                	   i = Integer.parseInt(hexpid,16);
                	}
                }
                log_.debug("findBGProcesses, dispatch process: " + str + ", pid: " + i + " [" + str2 + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

                bp = (OracleBackgroundProcess) ModelFactory.newInstance(OracleBackgroundProcess.class);
                bp.setParent(parent);
                bp.setName(str);
                bp.setDescription(str2);
                bp.setPid(i);
                v.add(bp);
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.20"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.12")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleBackgroundProcess[] oarray = new OracleBackgroundProcess[v.size()];
        return (OracleBackgroundProcess[]) v.toArray(oarray);
    }

    /**
     * Find all currently running user Oracle processes. These are any active
     * connections which are not marked as system background tasks. This may
     * include some system processes if they are not defined as background
     * tasks. Fetches the name and system pid. It's also possible to get the os &
     * oracle users, and plenty of of other info, if needed.
     */
    public OracleServerProcess[] findServerProcesses(OracleInstance parent, Connection conn) throws AgentException {
        OracleServerProcess sp;
        Vector v = new Vector(100);
        String str;
        int i;
        long l = 0;

        String query = "select p.program, p.spid FROM v$process p " + "WHERE background is null"; //$NON-NLS-1$ //$NON-NLS-2$

        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;
            OperatingSystem os = null;

            try {
                os = os_.getOperatingSystem();
            } catch (Exception oe) {
                log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.13")); //$NON-NLS-1$
            }

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                if(!isOpenVms_) {
                    i = rs.getInt(2);
                    l = i;
                } else {
                	// In OpenVms a pid is hex number
                	String hexpid = rs.getString(2);
                	if(hexpid != null && hexpid.length() > 0) {
                	   l = Long.parseLong(hexpid,16);
                	}
                }
                log_.debug("findServerProcesses, process: " + str + ", pid: " + l); //$NON-NLS-1$ //$NON-NLS-2$

                sp = (OracleServerProcess) ModelFactory.newInstance(OracleServerProcess.class);
                sp.setParent(os);
                sp.setName(str);
                sp.setPID(l);
                RuntimeProcess rp = os_.getProcessByPid(new Long(l));
                if (rp != null && rp.hasPorts()) {
                    sp.setPorts(rp.getPorts());
                    sp.setConnections(rp.getConnections());
                }
                v.add(sp);
            }
            stmt.close();

        } catch (Exception ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.20"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.14")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleServerProcess[] oarray = new OracleServerProcess[v.size()];
        return (OracleServerProcess[]) v.toArray(oarray);
    }

    /**
     * Finds the data files associated with the oracle database. This is where
     * all of the actual table data is stored.
     */
    public OracleDataFile[] findDataFiles(OracleDatabase parent, Connection conn) throws AgentException {
        OracleDataFile df;
        OracleTableSpace ts;
        Vector v = new Vector(100);
        String str;
        String str2;
        long i;

        String query = "select file_name, tablespace_name, bytes " + "from dba_data_files order by tablespace_name, file_name"; //$NON-NLS-1$ //$NON-NLS-2$

        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                str2 = rs.getString(2);
                i = rs.getLong(3);
                log_.debug("findDataFiles, file: " + str + ", size: " + i + ", table space: " + str2); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
                df = (OracleDataFile) ModelFactory.newInstance(OracleDataFile.class);
                df.setParent(parent);
                df.setName(str);
                df.setSize(i);
                ts = findSpaceByName(parent, str2);
                if (ts != null) {
                    df.setTableSpace(ts);
                }
                v.add(df);
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.21"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.15")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleDataFile[] oarray = new OracleDataFile[v.size()];
        return (OracleDataFile[]) v.toArray(oarray);
    }

    /**
     * Finds the control files associated with the oracle database. They store
     * backup, archive info and lots of other info.
     */
    public OracleControlFile[] findControlFiles(OracleDatabase parent, Connection conn) throws AgentException {
        OracleControlFile cf;
        Vector v = new Vector(100);
        String str;

        String query = "select name from v$controlfile"; //$NON-NLS-1$

        log_.debug(query);
        try {
            Statement stmt;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                str = rs.getString(1);
                log_.debug("findControlFiles, file: " + str); //$NON-NLS-1$
                cf = (OracleControlFile) ModelFactory.newInstance(OracleControlFile.class);
                cf.setParent(parent);
                cf.setName(str);
                v.add(cf);
            }
            stmt.close();

        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.22"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.16")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleControlFile[] oarray = new OracleControlFile[v.size()];
        return (OracleControlFile[]) v.toArray(oarray);
    }

    /**
     * Returns all entries in the v$sga System Global Area view. This table will
     * have an indeterminate number of entries describing the data buffers, redo
     * logs & shared pool. The init values will also describe the SGA.
     */
    public OracleSGAValue[] findSGAValues(OracleInstance parent, Connection conn) throws AgentException {
        OracleSGAValue sga;
        Vector v = new Vector(10);
        ResultSet rs = null;
        String str;
        Statement stmt;
        long i;

        String query = "select name, value from v$sga"; //$NON-NLS-1$

        try {
            log_.debug("findSGAValues about to execute:" + query); //$NON-NLS-1$
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
            while (rs.next()) {
                str = rs.getString(1);
                i = rs.getLong(2);
                log_.debug("findSGAValues found: " + str + " = " + i); //$NON-NLS-1$ //$NON-NLS-2$

                sga = (OracleSGAValue) ModelFactory.newInstance(OracleSGAValue.class);
                sga.setParent(parent);
                sga.setName(str);
                sga.setValue(i);

                v.add(sga);
            }

            stmt.close();

        } catch (SQLException ex) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.17"), ex); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.18")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleSGAValue[] oarray = new OracleSGAValue[v.size()];
        return (OracleSGAValue[]) v.toArray(oarray);
    }

    /**
     * Checks is extended feature is enabled by looking up collation.properties
     * file.
     */
    private boolean isSensorExtended() {
        return Props.getExtendedStatus();
    }

    /**
     * Queries the data dictionary table for host names that are used as an
     * connect string during creation of database links The host is later mapped
     * in tnsnames.ora file for retrieving ip, port & sid of the target
     * database, which is later drawn as transactional dependency.
     */

    private String[] getDBLinkData(Connection conn) throws AgentException {
        Vector v = new Vector(2);
        String query = "SELECT distinct host FROM dba_db_links"; //$NON-NLS-1$

        log_.debug(query);
        try {
            Statement stmt = null;
            ResultSet rs = null;

            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);

            while (rs.next()) {
                v.add(rs.getString(1));
            }

            stmt.close();
        } catch (SQLException ex) {
            log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.23"), ex); //$NON-NLS-1$
            log_.info(DiscoverAgentLocalizedMessages.getString("OracleAgent.I.4")); //$NON-NLS-1$
        }

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.19")); //$NON-NLS-1$
        }

        String[] sarr = new String[v.size()];
        return (String[]) v.toArray(sarr);
    }

    /**
     * Utility method that reads from a file (tnsnames.ora) and return the
     * result as a String
     * 
     * @param path
     *            URl of the <code>File </code> value
     * @return result a <code>String</code> value
     */
    private String getTNSContent(String oracle_home) {

        String tns_file = oracle_home.concat("/network/admin/").concat("tnsnames.ora"); //$NON-NLS-1$ //$NON-NLS-2$
        log_.debug("[getTNSContent] tns_file path/name : " + tns_file); //$NON-NLS-1$
        try {
            return returnLogicalContent(tns_file);
        } catch (Exception ex) {
            log_.debug("getTNSContent:Exception caught in Outer getCongifFile" + ex.toString()); //$NON-NLS-1$

            return null;
        }
    }

    /**
     * Persisting the host, port and sid read from tnsnames.ora file into
     * OracleDBLink model
     */
    private OracleDBLink[] findDBLinks(OracleDatabase parent, Connection conn, String oraSID) throws AgentException {
        OracleDBLink dbLink = null;
        String[] dbLinks = getDBLinkData(conn);

        if (null == dbLinks) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.20")); //$NON-NLS-1$
            return null;
        }

        Vector v = new Vector(dbLinks.length);
        storeDBLinks(getTNSContent(getOraHome()), dbLinks, v, dbLink, oraSID, parent);

        if (v.size() == 0) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.21")); //$NON-NLS-1$
        }

        // convert the vector to an array
        OracleDBLink[] oarray = new OracleDBLink[v.size()];
        return (OracleDBLink[]) v.toArray(oarray);
    }

    /**
     * 
     * @param tnsContent
     * @param dbLinks
     * @param v
     * @param dbLink
     * @param oraSID
     * @param parent
     */
    private void storeDBLinks(String tnsContent, String[] dbLinks, Vector v, OracleDBLink dbLink, String oraSID, OracleDatabase parent) {
        final String searchStr[] = { "HOST", "PORT", "SERVICE_NAME" }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
        String host = ""; //$NON-NLS-1$
        String ipaddr = ""; //$NON-NLS-1$
        String port = ""; //$NON-NLS-1$
        String sid = ""; //$NON-NLS-1$
        boolean status = false;
        boolean isLinkFound = false;
        boolean canStore = false;

        String[] strStore = getTNSValues(tnsContent);

        for (int i = 0; i < dbLinks.length; i++) {

            log_.debug("[findDBLinks] Finding value for... [" + i + "] : " + dbLinks[i]); //$NON-NLS-1$ //$NON-NLS-2$

            for (int j = 0; j < strStore.length; j++) {
                if (!status) {
                    if (dbLinks[i].equalsIgnoreCase(strStore[j])) { // NET SERVICE NAME
                        status = true;
                        isLinkFound = true;
                    }
                } else {
                    if (searchStr[0].equalsIgnoreCase(strStore[j])) { // HOST
                        host = strStore[j + 1];
                        try {
                            ipaddr = (!(IpUtils.validateIp(host)) ? InetAddress.getByName(host).getHostAddress() : host);
                            log_.debug("[findDBLinks - " + j + "] > Storing HOST for " + dbLinks[i] + " : [" + ipaddr + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$

                        } catch (UnknownHostException e) {
                            log_.error(DiscoverAgentLocalizedMessages.getFormattedString("OracleAgent.E.24", new Object[]{host, e})); //$NON-NLS-1$
                            status = false;
                            break;
                        }
                        continue;
                    }
                    if (searchStr[1].equalsIgnoreCase(strStore[j])) { // PORT
                        log_.debug("[findDBLinks - " + j + "] > Storing PORT for " + dbLinks[i] + " : [" + strStore[j + 1] + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
                        port = strStore[j + 1];
                        continue;
                    }
                    if (searchStr[2].equalsIgnoreCase(strStore[j])) { // SID
                        log_.debug("[findDBLinks - " + j + "] > Storing SID for " + dbLinks[i] + " : [" + strStore[j + 1] + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
                        sid = strStore[j + 1].toLowerCase();
                        status = false;
                        canStore = true;
                        break;
                    }
                } // End else
            } // End Inner For
            if (isLinkFound && canStore) {
                persistDBLinks(ipaddr, port, sid, oraSID, v, dbLink, parent);
                isLinkFound = false;
                canStore = false;
            }

        } // End Outer For
    }

    /**
     * 
     * @param ipAddr
     * @param port
     * @param sid
     * @param oraSID
     * @param v
     * @param dbLink
     * @param parent
     */
    private void persistDBLinks(String ipAddr, String port, String sid, String oraSID, Vector v, OracleDBLink dbLink, OracleDatabase parent) {
        if (sid.trim().length() > 0) {
            if (!(sid.equals(oraSID))) {
                dbLink = (OracleDBLink) ModelFactory.newInstance(OracleDBLink.class);

                dbLink.setParent(parent);
                dbLink.setIpAddress(ipAddr);
                dbLink.setPort(port);
                dbLink.setServiceName(sid);

                v.add(dbLink);
                log_.debug("[findDBLinks] - PERSISTING [" + oraSID + "] HOST : " + ipAddr + " , PORT : " + port + " , SID : " + sid //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
                        + " , Vector Size : " + v.size()); //$NON-NLS-1$
            }
        }
    }

    /**
     * Tokenizes the tnsnames.ora file content Stores and returns the tokens as
     * String array.
     */
    private String[] getTNSValues(String content) {
        if (content == null) {
            return new String[0];
        }
        StringTokenizer token = new StringTokenizer(content, "()="); //$NON-NLS-1$
        String[] storeStr = new String[token.countTokens()];
        int incr = 0;
        log_.debug("[getContent] Extracting tnsnames.ora file content..."); //$NON-NLS-1$
        while (token.hasMoreTokens()) {
            storeStr[incr] = token.nextToken().trim();
            log_.debug("Store Token [" + incr + "] : [" + storeStr[incr] + "]"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            incr++;
        } // End While
        return storeStr;
    }

    /**
     * Returns the tnsnames.ora file content as String
     */

    private String returnLogicalContent(String file) {
        try {
            String uri = URIUtils.newURI(primaryIp_.getDotNotation(), file).toString();
            log_.debug("[returnLogicalContent] Getting Logical Contents For URI : " + uri); //$NON-NLS-1$

            ConfigFile cf = (ConfigFile) os_.getLogicalContent(uri, Os.ALL_CONTENT);
            return cf.getContent();
        } catch (Exception ex) {
            log_.debug("[returnLogicalContent] Exception caught while processing the file " + file + " => " + ex.toString(), ex); //$NON-NLS-1$ //$NON-NLS-2$
            return null;
        }

    }

    /**
     * Returns discovered schema object defintions (metadata)
     * 
     * @param oinst
     * @param conn
     * @return
     * @throws AttributeNotSetException
     */

    private OracleSchemaDetail findSchemaObjects(OracleDatabase parent, OracleInstance oinst, Connection conn) throws AttributeNotSetException {

        OracleAgentDelegate metaAgent = new OracleAgentDelegate(conn, getModeType(), configuration_.getExtendedSchemaDiscovery().getSchemaObjects());
        Object metadata = metaAgent.discover();
        OracleSchemaDetail schemaContent = null;

        if (null != metadata) {
            try {
                schemaContent = (OracleSchemaDetail) ModelFactory.newInstance(OracleSchemaDetail.class);
                log_.debug("findSchemaObjects => Setting parent as database!"); //$NON-NLS-1$
                schemaContent.setParent(parent);
                schemaContent.setMetaData(metadata.toString());
                schemaContent.setChecksum(SecurityUtils.getMessageDigest(metadata, "MD5")); //$NON-NLS-1$
                schemaContent.setURI(URIUtils.newURI(com.collation.platform.model.topology.app.db.oracle.OracleDatabase.class,
                        primaryIp_.getDotNotation() + oinst.getHome() + "/" + oinst.getSID(), "metaData").toString()); //$NON-NLS-1$ //$NON-NLS-2$
            } catch (com.collation.platform.model.AttributeNotSetException ex) {
                log_.error(DiscoverAgentLocalizedMessages.getString("OracleAgent.E.25"), ex); //$NON-NLS-1$
            }
        } else {
            log_.info(DiscoverAgentLocalizedMessages.getString("OracleAgent.I.5")); //$NON-NLS-1$
        }

        return schemaContent;
    }

    private boolean isNameExists() throws AttributeNotSetException {
        return configuration_.getExtendedSchemaDiscovery().hasName();
    }

    private boolean isModeExists() throws AttributeNotSetException {
        return configuration_.getExtendedSchemaDiscovery().hasMode();
    }

    private int getModeType() throws AttributeNotSetException {

        if ((isNameExists()) && (isModeExists())) {
            try {
                return configuration_.getExtendedSchemaDiscovery().getMode();
            } catch (AttributeNotSetException e) {
                log_.warning(e);
            }
        }
        return OracleMetadataImpl.SHALLOW_MODE;
    }

    /**
     * Return the table space with the given name, as previously discovered and
     * stored in the database model object. This requires findTableSpaces to
     * have been run first.
     */
    private OracleTableSpace findSpaceByName(OracleDatabase db, String name) throws AgentException {
        OracleTableSpace retval = null;
        int i;

        OracleTableSpace[] spaces;
        if (name != null) {
            try {
                spaces = db.getTableSpaces();
                for (i = 0; i < spaces.length; i++) {
                    if (spaces[i].getName() != null) {
                        if (spaces[i].getName().compareTo(name) == 0) {
                            retval = spaces[i];
                            break;
                        }
                    }
                }
            } catch (AttributeNotSetException ex) {
                // not possible, but not fatal either
                log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.22")); //$NON-NLS-1$
            }
        }

        // it's ok for no match to be found, as some objects, like views,
        // / are not stored in a table space
        return retval;
    }

    /**
     * free all open connections
     */
    private void freeConnections() {

        for (Iterator conns = connPool.keySet().iterator(); conns.hasNext();) {
            String key = (String) conns.next(); // conn name
            Connection conn = (Connection) connPool.get(key);
            if (conn != null) {
                try {
                    conn.close();
                } catch (Exception ex) {
                    log_.debug("freeConnection, exception during close connection:", ex); //$NON-NLS-1$
                }
            }
        }
    }

    /**
     * Should generate a hashmap of (listener string, Array port list)
     */
    private synchronized HashMap getPortList() throws AgentException {

        HashMap portMap = new HashMap();
        String pscmd = null;
        String psres = null;

        try {
            BindAddress[] listeningPortList = oraSeed_.getProcess().getPorts();
            Long pid = new Long(oraSeed_.getProcess().getPID());
            String tns = oraSeed_.getProcess().getCmdLine();

            log_.debug("Found listener: " + tns + ", pid: " + pid); //$NON-NLS-1$ //$NON-NLS-2$
            ArrayList plist = (ArrayList) portMap.get(tns);
            if (plist == null) {
                plist = new ArrayList();
                portMap.put(tns, plist);
            }

            RuntimeProcess rp = oraSeed_.getProcess();
            if (rp != null && rp.hasPorts()) {

                BindAddress[] ba = rp.getPorts();
                for (int i = 0; i < ba.length; i++) {
                    BindAddress b = ba[i];
                    if (b != null) {
                        int port = b.getPortNumber();
                        String ip = null;
                        try {
                            ip = b.getIpAddress().toString();
                        } catch (Exception ex) {
                        }
                        if (port == 8080) {
                            log_.debug("Ignoring open port for listener: " + tns + ", port: " + port); //$NON-NLS-1$ //$NON-NLS-2$
                            continue;
                        }
                        plist.add(b);
                        log_.debug("Found open port for listener: " + tns + "(" + pid + "), port: " + ip + ":" + port); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
                    }
                } // for each bind address
            }
        } catch (Exception ex) {
            throw new AgentException(DiscoverAgentLocalizedMessages.class, "OracleAgent.E.26", ex); //$NON-NLS-1$
        }

        return portMap;
    }

    public void discoverAppDescriptors() {
        try {
            if (result_ == null) {
                return;
            }
            log_.debug("discoverAppDescriptors()"); //$NON-NLS-1$
            String defaultAppDescriptorDir_ = Props.getOracleAppDescriptorDefaultLocation();
            log_.debug("discoverAppDescriptors() - default app descriptor dir was overridden to " + defaultAppDescriptorDir_); //$NON-NLS-1$
            OracleInstance[] insts = result_.getOracleInstances();
            if (insts != null) {
                for (int j = 0; j < insts.length; j++) {
                    if (insts[j] != null) {
                        String defaultDir = null;
                        try {
                        	if (!isOpenVms_) {
                                if (defaultAppDescriptorDir_ != null && !defaultAppDescriptorDir_.trim().equals("")) { //$NON-NLS-1$
                                    defaultDir = defaultAppDescriptorDir_ + "/" + insts[j].getSID() + "/" //$NON-NLS-1$ //$NON-NLS-2$
                                        + AppDescriptorAgentDelegate.DEFAULT_APP_DESCRIPTOR_DIRNAME;
                                } else {
                                    defaultDir = insts[j].getHome() + "/" + insts[j].getSID() + "/" //$NON-NLS-1$ //$NON-NLS-2$
                                        + AppDescriptorAgentDelegate.DEFAULT_APP_DESCRIPTOR_DIRNAME;
                                }
                            } else {
                            	if (defaultAppDescriptorDir_ != null && !defaultAppDescriptorDir_.trim().equals("")) { //$NON-NLS-1$
                                    defaultDir = defaultAppDescriptorDir_ + "." + insts[j].getSID() + "." //$NON-NLS-1$ //$NON-NLS-2$
                                            + AppDescriptorAgentDelegate.DEFAULT_APP_DESCRIPTOR_DIRNAME + "]"; //$NON-NLS-1$
                                } else {
                                	String homeDir = insts[j].getHome();
                                    defaultDir = "/" + homeDir.substring(0,homeDir.length() - 1) + "."   //$NON-NLS-1$ //$NON-NLS-2$
                                            + insts[j].getSID() + "." //$NON-NLS-1$
                                            + AppDescriptorAgentDelegate.DEFAULT_APP_DESCRIPTOR_DIRNAME + "]"; //$NON-NLS-1$
                                }
                            	log_.debug("discoverAppDescriptors: defaultDir set to: " + defaultDir); //$NON-NLS-1$
                            }
                        } catch (AttributeNotSetException ex) {
                            defaultDir = null;
                        }
                        AppDescriptorAgentDelegate.discover(result_, insts[j], defaultDir);
                    }
                }
            }
        } catch (Exception ignore) {
            log_.warning(DiscoverAgentLocalizedMessages.getString("OracleAgent.W.23"), ignore); //$NON-NLS-1$

        }
    }

    private boolean isInDirectory(String home, String path) {
        if (home == null || path == null) {
            return false;
        }
        int i = path.indexOf(home);
        if (i != 0) {
            return false;
        }
        if (path.charAt(home.length()) != '/') {
            return false;
        }
        return true;
    }

    /**
     * Eliminates some redundant calls to "ps". The real fix is to parse the
     * output in Java rather than do a couple of greps.
     */
    private String getPsCache() throws com.collation.platform.os.OsException, com.collation.platform.session.SessionClientException,
            java.lang.InterruptedException {
        String res = null;
        synchronized (psCache_) {
            res = (String) psCache_.get(hostName_);

            if (res == null) {
                /*
                 * Removed used at Fidelity, where "ps axww" would hang. Perhaps
                 * it's an SSH bug. More testing required. if (os_ instanceof
                 * SolarisOs) { pscmd = "/bin/ps -ef"; } else { pscmd =
                 * os_.getCommandPsDetails(); // "ps axww" }
                 */
                OsCmd pscmd = os_.getCommandPsDetails(); // "ps axww"

                //pscmd.append(" | grep ora_pmon | grep -v grep");
                res = pscmd.execute();
		String[] procs = res.split("\n"); //$NON-NLS-1$
		StringBuffer sb = new StringBuffer();
		for (int i=0; i <procs.length; i++){
		    if (isWindows_){
			if (procs[i].toLowerCase().indexOf("oracle.exe") > 0){ //$NON-NLS-1$
				//log_.debug("Proc Filter adding " + procs[i]);
				sb.append(procs[i] + "\n"); //$NON-NLS-1$
			}
		    }
		    else if (isOpenVms_) {
		    	// OpenVMS tasks are ORA_<sid>_PMON
		    	if (procs[i] != null) {
		    		String[] procParts = procs[i].split(" "); //$NON-NLS-1$
		    		// The task name is the 2nd word
		    		if (procParts.length > 1) {
		    			String openVmsTaskName = procParts[1].toUpperCase();
		    			if (openVmsTaskName.startsWith("ORA_") &&  //$NON-NLS-1$
		    				openVmsTaskName.endsWith("_PMON")) { //$NON-NLS-1$
		    				sb.append(procs[i] + "\n"); //$NON-NLS-1$
		    			}
		    		}
		    	}
		    } else {
			if (procs[i].indexOf("ora_pmon") > 0 ){ //$NON-NLS-1$
				sb.append(procs[i] + "\n"); //$NON-NLS-1$
			}
		    }
		}
		res = sb.toString();			
                psCache_.put(hostName_, res);
            }
        }
        return res;
    }

    /**
     * @return Returns the oraHome.
     */
    private String getOraHome() {
        return oraHome;
    }

    /**
     * @param oraHome
     *            The oraHome to set.
     */
    private void setOraHome(String oraHome) {
        this.oraHome = oraHome;
    }

    /**
     * Verify that a connection to an IP address via a port can be made. This
     * will prevent infinite hangs
     * 
     * @return
     */
    public boolean testPort(String ip, int port) {
        //First see if it is known bad already
        if (badEndpoints_.containsKey(ip + ":" + port)){ //$NON-NLS-1$
            //this one is bad
            log_.debug("testPort found known bad endpoint: " + ip + ":" + port);//$NON-NLS-1$ //$NON-NLS-2$
            return false;
        }

        try {
            InetSocketAddress isock = new InetSocketAddress(ip, port);
            Socket sk = new Socket();
            sk.setSoLinger(false, 0);
            sk.connect(isock, 5000);
            sk.close();
            return true;
        } catch (Exception e) {
            log_.debug("testPort(" + ip + ":" + port + "): " + e.toString()); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            badEndpoints_.put(ip + ":" + port, "badport"); //$NON-NLS-1$ //$NON-NLS-2$
            log_.debug("testPort added " + ip + ":" + port + " to badEndpoints."); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            return false;
        }
    }

    /**
     * Return all ip interfaces which this bind address listens on.
     */
    private String[] getAddrs(BindAddress ba) {
        ArrayList list = new ArrayList();
        try {
            String addr = ba.getIpAddress().getDotNotation();
            //if the bind address is 0.0.0.0, then get the ip addresses
            //from the box, otherwise, just use the bind addresss.
            if (ba.getIpAddress().getDotNotation().equals("0.0.0.0")) { //$NON-NLS-1$
                ArrayList list2 = os_.getInterfaces();
                for (int i = 0; i < list2.size(); i++) {
                    list.add(list2.get(i));
                }
            }
            else{
                list.add(addr);
            }
        } catch (Exception e) {
            log_.debug("getAddrs, unable to get address from: " + ba + ", " + e); //$NON-NLS-1$ //$NON-NLS-2$
        }

        String[] oarray = new String[list.size()];
        return (String[]) list.toArray(oarray);
    }

    /**
     * @see com.collation.platform.configuration.Configurable#getConfigurationClass()
     */
    public Class getConfigurationClass() {
        return OracleAgentConfiguration.class;
    }

    /**
     * @see com.collation.platform.configuration.Configurable#getConfiguration()
     */
    public Configuration getConfiguration() {
        return configuration_;
    }

    /**
     * @see com.collation.platform.configuration.Configurable#setConfiguration(com.collation.platform.model.Configuration)
     */
    public void setConfiguration(Configuration configuration) throws ConfigurationException {
        configuration_ = (OracleAgentConfiguration) configuration;
    }

    public class OracleJDBCConnector extends Thread{

        Connection conn_ = null;
        Exception ex_ = null;
        Driver d_; 
        String connStr_;
        Properties authProps_;
	private Logger log_ = null;
 
        public OracleJDBCConnector(Driver driver, String url, Properties auth, Logger logger){
            super();
	    setName("OracleJDBCConnector");  //$NON-NLS-1$ 
            setDaemon(true);
            d_ = driver;
            connStr_ = url;
            authProps_ = auth;
            log_ = logger;
        }
  
        public void run(){
            try{
                synchronized (this) {
                    log_.debug("OracleJDBCConnector about to connect to " + connStr_);  //$NON-NLS-1$
                    conn_ = d_.connect(connStr_,authProps_);
                    log_.debug("OracleJDBCConnector connection establised, finishing");  //$NON-NLS-1$
                }
            }
            catch (Exception ex){
                ex_ = ex;
            }
        }
   
        public Connection getConnection(){
            return conn_;
        }

        public Exception getException(){
            return ex_;
        }
    } 
}
